# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import unittest
from unittest import mock

import pkg_resources


def require_gui(func):
    """ Decorator for tests that require a non-null GUI toolkit.
    """
    # Defer GUI import side-effect.
    # The toolkit is not resolved until we import pyface.api
    from pyface.api import GUI
    try:
        GUI()
    except NotImplementedError:
        return unittest.skip("No GUI available.")(func)
    return func


def create_iter_entry_points(fake_entry_points):
    """ Return a new callable that mocks the interface of
    pkg_resources.iter_entry_points with the given entry point data.

    Currently the behaviour of a given non-None value for ``name`` is not
    supported.

    Parameters
    ----------
    fake_entry_points : dict(str, dict)
        Mapping from distribution names to entry point definitions.
    """

    def iter_entry_points(group, name=None):

        if name is not None:
            # Mocking this logic is not currently needed.
            raise ValueError(
                "Currently the test code does not mock the behavior when name "
                "is not None."
            )

        for dist_name, group_to_entry_points in fake_entry_points.items():
            for entry_point_text in group_to_entry_points.get(group, []):
                entry_point = pkg_resources.EntryPoint.parse(
                    entry_point_text,
                    dist=pkg_resources.get_distribution(dist_name),
                )
                yield entry_point

    return iter_entry_points


def mock_iter_entry_points(fake_entry_points):
    """ Mock pkg_resources.iter_entry_points with the given entry point data.

    Currently the behaviour of a given non-None value for ``name`` is not
    supported.

    Parameters
    ----------
    fake_entry_points : dict(str, dict)
        Mapping from distribution names to entry point definitions.
    """
    return mock.patch.object(
        pkg_resources, "iter_entry_points",
        create_iter_entry_points(fake_entry_points)
    )



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import os
import tempfile
import textwrap
import unittest
from xml.etree import ElementTree as ET

from traitsui.api import Handler, UI, UIInfo

from etsdemo.app import (
    Demo,
    DemoPath,
    DemoVirtualDirectory,
    extract_docstring_from_source,
    parse_source,
    next_tool,
    previous_tool,
    parent_tool,
)

HTML_NS_PREFIX = "{http://www.w3.org/1999/xhtml}"


def get_html_tag(tag):
    return HTML_NS_PREFIX + tag


def get_action_enabled(action, model_view):
    """ Helper funciton to return if a tool is enabled.

    Parameters
    ----------
    action : Action
    model_view : ModelView
    """
    context = model_view.trait_get()
    context.update(model_view.trait_context())
    return eval(
        compile(action.enabled_when, "<string>", "eval"),
        {},
        context,
    )


class TestDemo(unittest.TestCase):
    """ Test actions on the Demo class. """

    def test_demo_previous_button_default(self):
        # Make sure the previous button behaves in the default state.
        # This ensures the enable flag can be computed (instead of crashing),
        # and the action allowed by it does not crash.
        demo = Demo(model=DemoPath())
        if get_action_enabled(previous_tool, demo):
            demo.perform(None, next_tool, None)

    def test_demo_next_button_default(self):
        # Make sure the next button behaves in the default state.
        # This ensures the enable flag can be computed (instead of crashing),
        # and the action allowed by it does not crash.
        demo = Demo(model=DemoPath())
        if get_action_enabled(next_tool, demo):
            demo.perform(None, next_tool, None)

    def test_demo_parent_button_default(self):
        # Make sure the parent button behaves in the default state.
        # This ensures the enable flag can be computed (instead of crashing),
        # and the action allowed by it does not crash.
        demo = Demo(model=DemoPath())
        if get_action_enabled(parent_tool, demo):
            demo.perform(None, parent_tool, None)

    def test_demo_init_set_children(self):
        # Test if there are children, the first one will be selected.
        resources = [DemoPath(), DemoPath()]
        model = DemoVirtualDirectory(resources=resources)
        demo = Demo(model=model)
        demo.selected_node = None

        # when
        info = UIInfo(ui=UI(handler=Handler()))
        demo.init(info)

        # then
        self.assertIs(demo.selected_node, resources[0])

    def test_demo_init_no_children_to_be_set(self):
        # Test if there are no children, nothing is selected.
        model = DemoVirtualDirectory(resources=[])
        demo = Demo(model=model)
        demo.selected_node = None

        # when
        info = UIInfo(ui=UI(handler=Handler()))
        demo.init(info)

        # then
        self.assertIsNone(demo.selected_node)


class TestDemoPathDescription(unittest.TestCase):
    """ Test ``DemoPath.description`` """

    def test_description_with_empty_directory(self):
        # If the directory is empty, the content of the description should
        # be empty.
        with tempfile.TemporaryDirectory() as directory:
            model = DemoPath(
                name=directory,
            )

            tree = ET.fromstring(model.description)
        body_node = next(tree.iter(get_html_tag("body")))
        div_node, = list(body_node)
        self.assertEqual(list(div_node), [])

    def test_use_index_rst(self):
        with tempfile.TemporaryDirectory() as directory:
            index_rst = os.path.join(directory, "index.rst")
            with open(index_rst, "w", encoding="utf-8") as f:
                f.write(".. image:: any_image.jpg\n")

            model = DemoPath(
                name=directory,
            )

            tree = ET.fromstring(model.description)
        img_node = next(tree.iter(get_html_tag("img")))
        self.assertEqual(img_node.attrib["src"], "any_image.jpg")

    def test_description_use_css(self):
        with tempfile.TemporaryDirectory() as directory:
            model = DemoPath(
                name=directory,
                css_filename="default.css",
            )

            tree = ET.fromstring(model.description)

        link_node = next(tree.iter(get_html_tag("link")))
        self.assertEqual(link_node.attrib["href"], "default.css")


class TestDemoPathInitLocals(unittest.TestCase):
    """ Test ``DemoPath.init_dic`` """

    def test_init_dict_with_empty_directory(self):
        with tempfile.TemporaryDirectory() as directory:
            model = DemoPath(
                name=directory,
            )
            # traits api is still imported.
            init_dic = model.init_dic
            self.assertIsInstance(init_dic, dict)
            self.assertIn("HasTraits", init_dic)

    def test_init_dict_with_init_py(self):
        with tempfile.TemporaryDirectory() as directory:
            init_py = os.path.join(directory, "__init__.py")
            with open(init_py, "w", encoding="utf-8") as f:
                f.write("a = 1\n")
            model = DemoPath(
                name=directory,
            )
            self.assertIn("a", model.init_dic)
            self.assertEqual(model.init_dic["a"], 1)
            # traits api is still imported.
            self.assertIn("HasTraits", model.init_dic)


class TestDemoPathChildren(unittest.TestCase):
    """ Integration test with DemoPath and its children
    """

    def test_init_dict_used_by_children(self):
        # Test the __init__.py in a directory (if exists) is visible
        # by the example scripts in that folder.
        # Not sure if this is really needed, but this is an existing feature.
        with tempfile.TemporaryDirectory() as directory:

            init_content = "CONSTANT = 'HELLO'"
            example_content = textwrap.dedent(
                """
                from . import CONSTANT
                from . import HasTraits
                """
            )
            subdir = os.path.join(directory, "Examples")
            os.makedirs(subdir)
            init_py = os.path.join(subdir, "__init__.py")
            with open(init_py, "w", encoding="utf-8") as f:
                f.write(init_content)
            example_py = os.path.join(subdir, "example.py")
            with open(example_py, "w", encoding="utf-8") as f:
                f.write(example_content)

            model = DemoPath(
                name=directory,
                use_files=False,
            )

            # sanity check:
            # This is one subdirectory
            children = model.get_children()
            self.assertEqual(len(children), 1)

            # In that subdirectory, there is one Python file that is not
            # __init__.py
            subdir_node, = children
            file_nodes = subdir_node.get_children()
            self.assertEqual(len(file_nodes), 1)
            example, = file_nodes
            self.assertEqual(example.name, "example.py")

            # This is the test objective: The __init__.py and traits api are
            # loaded and accessible by the example script.
            # Try running the code
            example.run_code()
            self.assertIn("CONSTANT", example.locals)
            self.assertIn("HasTraits", example.locals)


class TestDemoVirtualDirectory(unittest.TestCase):

    def test_no_resources(self):
        node = DemoVirtualDirectory(resources=[])
        self.assertFalse(node.tno_has_children(None))
        self.assertEqual(node.tno_get_children(None), [])
        self.assertTrue(node.tno_allows_children(None))

    def test_some_resources(self):
        node = DemoVirtualDirectory(resources=[DemoPath()])
        self.assertTrue(node.tno_has_children(None))
        self.assertEqual(node.tno_get_children(None), node.resources)
        self.assertTrue(node.tno_allows_children(None))


class TestParseSource(unittest.TestCase):

    def test_extract_docstring_from_source(self):
        source_code = b""
        with self.assertRaises(TypeError):
            docstring, source = extract_docstring_from_source(source_code)

        source_code = u""
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual((u"", u""), (docstring, source))

        source_code = u'''""" Module description """\nx=1\ny=2'''
        docstring, source = extract_docstring_from_source(source_code)
        expected = (u" Module description ", "x=1\ny=2")
        self.assertEqual(expected, (docstring, source))

    def test_parse_source(self):
        docstring, source = parse_source("non-existent-file.<>|:")
        self.assertIn("Sorry, something went wrong.", docstring)

    def test_simple_source(self):
        source_code = "\n".join(['"""', "Docstring", '"""', "a = 1"])
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual(docstring, "\nDocstring\n")
        self.assertEqual(source, "a = 1")

    def test_alternate_quotes(self):
        source_code = "\n".join(["'''", "Docstring", "'''", "a = 1"])
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual(docstring, "\nDocstring\n")
        self.assertEqual(source, "a = 1")

    def test_string_in_source(self):
        source_code = "\n".join(
            ['"""', "Docstring", '"""', '"string in source"', "a = 1"]
        )
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual(docstring, "\nDocstring\n")
        self.assertEqual(source, "\n".join(['"string in source"', "a = 1"]))

    def test_string_in_docstring(self):
        source_code = "\n".join(
            ['"""', "Docstring", '"string in docstring"', '"""', "a = 1"]
        )
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual(
            docstring,
            "\n".join(["", "Docstring", '"string in docstring"', ""]),
        )
        self.assertEqual(source, "\n".join(["a = 1"]))

    def test_ignore_class_docstring(self):
        source_code = "\n".join(["class Foo:", '    """Class docstring"""'])
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual(docstring, "")
        self.assertEqual(
            source, "\n".join(["class Foo:", '    """Class docstring"""'])
        )

    def test_ignore_starting_comment(self):
        source_code = "\n".join(
            [
                "# Copyright notice.",
                "# Something about the author.",
                '"""',
                "Docstring",
                '"""',
                "a = 1",
            ]
        )
        docstring, source = extract_docstring_from_source(source_code)
        self.assertEqual(docstring, "\nDocstring\n")
        self.assertEqual(
            source,
            "\n".join(
                [
                    "# Copyright notice.",
                    "# Something about the author.",
                    "a = 1",
                ]
            ),
        )



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import io
import tempfile
import unittest
from unittest import mock

from pyface.api import GUI

from etsdemo.app import Demo
from etsdemo import main as main_module
from etsdemo.tests.testing import (
    mock_iter_entry_points,
    require_gui,
)


def fake_configure_traits(instance):
    ui = instance.edit_traits()
    try:
        GUI.process_events()
    finally:
        try:
            ui.dispose()
        finally:
            GUI.process_events()


def mock_demo_launch():
    """ Mock Demo.configure_traits so that we can open and then close the UI
    for testing the main function.
    """
    return mock.patch.object(Demo, "configure_traits", fake_configure_traits)


class TestMain(unittest.TestCase):
    """ Test main function."""

    @require_gui
    def test_main(self):
        # Main function must be launchable even if there are no data available.
        # In normal running situation, no loggings should be emitted
        argv = ["etsdemo"]
        mocked_io = io.StringIO()
        with mock_iter_entry_points({}), \
                mock_demo_launch(), \
                mocked_io, \
                mock.patch("sys.stdout", mocked_io), \
                mock.patch("sys.stderr", mocked_io), \
                mock.patch("sys.argv", argv):
            main_module.main()
            console_output = mocked_io.getvalue()

        self.assertNotIn("Found 0 resource(s).", console_output)

    @require_gui
    def test_main_with_log(self):
        # Test logging configuration with the main function.
        argv = ["etsdemo", "-v"]
        mocked_io = io.StringIO()
        with mock_iter_entry_points({}), \
                mock_demo_launch(), \
                mocked_io, \
                mock.patch("sys.stdout", mocked_io), \
                mock.patch("sys.stderr", mocked_io), \
                mock.patch("sys.argv", argv):
            main_module.main()
            console_output = mocked_io.getvalue()

        self.assertIn("Found 0 resource(s).", console_output)


class TestCreateDemo(unittest.TestCase):
    """ Test _create_demo """

    def test_create_demo_default_entry_points(self):
        # This does not require GUI and it should not fail.
        # This will load the entry points contributed by the dependencies
        # of etsdemo.
        main_module._create_demo()

    def test_create_demo_with_specific_info(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            infos = [
                {
                    "version": 1,
                    "name": "Pretty Demo",
                    "root": temp_dir,
                },
                {
                    "version": 1,
                    "name": "Delicious Demo",
                    "root": temp_dir,
                }
            ]
            demo = main_module._create_demo(infos)
        children = demo.model.get_children()
        self.assertEqual(len(children), 2)
        child1, child2 = children
        self.assertEqual(child1.nice_name, "Pretty Demo")
        self.assertEqual(child2.nice_name, "Delicious Demo")

    def test_create_demo_with_title(self):
        demo = main_module._create_demo([], "Nice Title")
        self.assertEqual(demo.title, "Nice Title")



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Test the EAM app metadata.
"""

import os
import subprocess
import unittest

try:
    import eam   # noqa: #F401
except ImportError:
    EAM_EXISTS = False
else:
    EAM_EXISTS = True

from etsdemo.info import info


class TestInfoForEAM(unittest.TestCase):

    @unittest.skipUnless(EAM_EXISTS, "eam is not available.")
    def test_etsdemo_info(self):
        output = subprocess.check_output(
            ["eam", "info"],
        )
        output = output.decode("utf-8")
        self.assertIn("etsdemo", output)

    def test_etsdemo_icon_exists(self):
        metadata = info()
        command, = metadata["commands"]
        icon = command["icon"]
        self.assertTrue(
            os.path.exists(icon), "Expected icon file to exist. Not found."
        )



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!
"""
Version information for this distribution.
This file is autogenerated by the setup.py script.
"""
#: The full version of the package, including a development suffix
#: for unreleased versions of the package.
version = "0.1.0"
#: The Git revision from which this release was made.
git_revision = "8669b51b09e00b4c6b6bd0e9cf8ff83ff96a0d74"



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import pkg_resources

from etsdemo.version import version

ICON = pkg_resources.resource_filename("etsdemo", "images/enthought-icon.png")


def info():
    """ Provides information to the "eam" package. """
    return {
        "name": "ETS Demo",
        "description": "Application for browsing ETS examples and demos",
        "license": "BSD",
        "copyright": "(c) 2020 Enthought",
        "version": version,
        "schema_version": 2,
        "commands": [
            {
                "name": "ETS Demo",
                "command": "etsdemo",
                "icon": ICON,
            },
        ],
    }



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" Module for loading example data files contributed by other distributions
via entry points.
"""

import logging

import pkg_resources

from traits.api import Directory, HasTraits, Range, Str, TraitError

from etsdemo.app import DemoPath, DemoVirtualDirectory


logger = logging.getLogger(__name__)

#: The entry point group name for other packages to contribute
#: data files.
ENTRY_POINT_GROUP = "etsdemo_data"


def get_responses():
    """ Load entry points and return the responses.

    Version 1 response should have the following format:
    {
        "version": 1,
        "name": <str>,
        "root": <str>,
    }

    Required keys:
    ``version``: int
        The version for the format of this response
    ``name``: str
        The main display name for the group of resources.
    ``root``: str
        The root directory where resource files are stored.

    Returns
    -------
    responses: list of dict
        Responses loaded from the entry points.
        Note that the content has not been sanitized nor validated.
    """
    request = {}
    responses = []
    for entry_point in pkg_resources.iter_entry_points(ENTRY_POINT_GROUP):
        try:
            loader = entry_point.load()
        except ImportError:
            logger.exception("Failed to load entry point %r", entry_point)
            continue

        # `request` is currently a placeholder for future extension of the API
        # without having to change the function signature.
        try:
            response = loader(request)
        except Exception:
            logger.exception("Failed to obtain data from %r", entry_point)
            continue
        responses.append(response)
    return responses


def response_to_node(response):
    """ Convert a response to an instance of DemoTreeNodeObject

    Parameters
    ----------
    response : any
        Response obtained from an entry point.
        It is expected to be a dict, but there is no guarantee so it can
        be anything.

    Returns
    -------
    node : DemoTreeNodeObject
    """

    # Still try to get the name if we can
    try:
        name = response["name"]
    except (TypeError, KeyError):
        name = "(Empty)"

    try:
        normalized = _Response(
            version=response["version"],
            name=response["name"],
            root=response["root"],
        )
    except (TypeError, KeyError, TraitError):
        logger.exception("Failed to load response: %r", response)
        return DemoVirtualDirectory(
            nice_name=str(name),
            resources=[],
            description="Unable to load data.",
        )

    return normalized.to_node()


class _Response(HasTraits):
    """ Object for normalizing the validating responses returned by entry
    points.
    """

    #: Version
    version = Range(low=0, high=1)

    #: Display name
    name = Str()

    #: Root directory for any nested data files.
    root = Directory(exists=True)

    def to_node(self):
        """ Return an instance of DemoTreeNodeObject from this response.
        """
        return DemoPath(
            nice_name=self.name,
            name=self.root,
        )



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

import os
import tempfile
import unittest

from etsdemo.loader import (
    __name__ as LOGGER_NAME,
    get_responses,
    response_to_node,
)
from etsdemo.tests.testing import mock_iter_entry_points


def amazing_demo_response(request):
    return {
        "version": 1,
        "name": "Amazing Demo",
        "root": "",
    }


def wonderful_demo_response(request):
    return {
        "version": 1,
        "name": "Wonderful Demo",
        "root": "",
    }


def misbehaving_demo_response(request):
    raise ZeroDivisionError()


# Data to mock the entry points for testing.
# Mapping from distribution names to entry points
GOOD_ENTRY_POINTS = {
    "traitsui": {
        "etsdemo_data": [
            "amazing_demo={}:amazing_demo_response".format(__name__),
            "wonderful_demo={}:wonderful_demo_response".format(__name__),
        ],
    },
}


class TestResponseToNode(unittest.TestCase):
    """ Test conversion from response to node. """

    def test_good_response_to_node(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            with open(os.path.join(temp_dir, "index.rst"), "w"):
                pass
            response = {
                "version": 1,
                "name": "Amazing Demo",
                "root": temp_dir,
            }
            resource = response_to_node(response)
            self.assertEqual(resource.name, temp_dir)
            self.assertEqual(resource.nice_name, "Amazing Demo")
            self.assertTrue(resource.has_children())

    def test_good_response_but_nonexisting_root_to_node(self):
        # If the response refers to a nonexisting root path,
        # it is detected and replaced with a placeholder.
        response = {
            "version": 1,
            "name": "Amazing Demo",
            "root": "I_do_not_exist",
        }
        with self.assertLogs(LOGGER_NAME) as watcher:
            resource = response_to_node(response)
        self.assertEqual(resource.nice_name, "Amazing Demo")
        self.assertFalse(resource.has_children())
        self.assertIn(
            "Unable to load data.", resource.description,
        )
        log_content, = watcher.output
        self.assertIn("TraitError", log_content)

    def test_bad_response_replaced(self):
        # If the response is badly formatted, replace with a placeholder.
        response = {}
        with self.assertLogs(LOGGER_NAME) as watcher:
            resource = response_to_node(response)

        self.assertFalse(resource.has_children())
        self.assertEqual(resource.nice_name, "(Empty)")
        self.assertIn(
            "Unable to load data.",
            resource.description
        )
        log_content, = watcher.output
        self.assertIn("KeyError", log_content)

    def test_bad_response_type_error(self):
        bad_values = [
            None,
            "1",
            1,
            (),
        ]
        for bad_value in bad_values:
            with self.subTest(bad_value=bad_value):
                with self.assertLogs(LOGGER_NAME) as watcher:
                    resource = response_to_node(bad_value)

                self.assertFalse(resource.has_children())
                self.assertEqual(resource.nice_name, "(Empty)")
                self.assertIn(
                    "Unable to load data.",
                    resource.description
                )
                log_content, = watcher.output
                self.assertIn("TypeError", log_content)

    def test_bad_response_missing_name(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            response = {
                "version": 1,
                "root": temp_dir,
            }
            with self.assertLogs(LOGGER_NAME) as watcher:
                resource = response_to_node(response)

            self.assertFalse(resource.has_children())
            self.assertEqual(resource.nice_name, "(Empty)")
            log_content, = watcher.output
            self.assertIn("KeyError: \'name\'", log_content)

    def test_bad_response_missing_version(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            response = {
                "name": "Name",
                "root": temp_dir,
            }
            with self.assertLogs(LOGGER_NAME) as watcher:
                resource = response_to_node(response)

            self.assertFalse(resource.has_children())
            log_content, = watcher.output
            self.assertIn("KeyError: \'version\'", log_content)

    def test_bad_response_bad_version(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            response = {
                "version": 2,
                "name": "Name",
                "root": temp_dir,
            }
            with self.assertLogs(LOGGER_NAME) as watcher:
                resource = response_to_node(response)

            self.assertFalse(resource.has_children())
            log_content, = watcher.output
            self.assertIn("TraitError", log_content)

    def test_bad_response_bad_name_type(self):
        with tempfile.TemporaryDirectory() as temp_dir:
            response = {
                "version": 1,
                "name": 1,
                "root": temp_dir,
            }
            with self.assertLogs(LOGGER_NAME) as watcher:
                resource = response_to_node(response)

            self.assertFalse(resource.has_children())
            log_content, = watcher.output
            self.assertIn("TraitError", log_content)


class TestGetResponse(unittest.TestCase):
    """ Test the routine to obtain resources from entry points."""

    def test_load_good_entry_points(self):
        with mock_iter_entry_points(GOOD_ENTRY_POINTS):
            resources = get_responses()
        expected = [
            amazing_demo_response({}),
            wonderful_demo_response({}),
        ]
        self.assertEqual(resources, expected)

    def test_load_errored_entry_points(self):
        # entry points are contributed by other packages, and running them
        # may fail.
        error_entry_point = {
            "traitsui": {
                "etsdemo_data": [
                    "error_demo={}:misbehaving_demo_response".format(__name__),
                    "amazing_demo={}:amazing_demo_response".format(__name__),
                ],
            },
        }
        with mock_iter_entry_points(error_entry_point):
            with self.assertLogs(LOGGER_NAME) as watcher:
                resources = get_responses()

        self.assertEqual(len(resources), 1)
        log_output, = watcher.output
        self.assertIn("Failed to obtain data from", log_output)

    def test_load_import_error(self):
        # entry points are contributed by other packages, and loading them
        # may fail. In this test, the path is wrong and results in ImportError.
        error_entry_point = {
            "traitsui": {
                "etsdemo_data": [
                    "missing_demo={}:nonexisting_function".format(__name__),
                    "amazing_demo={}:amazing_demo_response".format(__name__),
                ],
            },
        }
        with mock_iter_entry_points(error_entry_point):
            with self.assertLogs(LOGGER_NAME) as watcher:
                resources = get_responses()

        self.assertEqual(len(resources), 1)
        log_output, = watcher.output
        self.assertIn("Failed to load entry point", log_output)



# (C) Copyright 2020 Enthought, Inc., Austin, TX
# All rights reserved.
#
# This software is provided without warranty under the terms of the BSD
# license included in LICENSE.txt and may be redistributed only under
# the conditions described in the aforementioned license. The license
# is also available online at http://www.enthought.com/licenses/BSD.txt
#
# Thanks for using Enthought open source!

""" The main function for launching the demo application.
"""
import argparse
import contextlib
import logging

from etsdemo.app import Demo, DemoVirtualDirectory
from etsdemo.loader import get_responses, response_to_node


#: Logging constants
_LOG_FORMAT = "%(asctime)s %(levelname)-8.8s [%(name)s:%(lineno)s] %(message)s"
logger = logging.getLogger(__name__)

#: Default application title
_TITLE = "Enthought Tool Suite"

_THIS_PACKAGE, _ = __name__.split(".", 1)


@contextlib.contextmanager
def _set_logger(logger, level):
    """ Context manager for setting logging.

    Parameters
    ----------
    logger : logging.Logger
        Logger to be configured.
    level : int
        Logging level to set.
    """
    formatter = logging.Formatter(_LOG_FORMAT)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    handler.setFormatter(formatter)

    original_level = logger.level
    logger.setLevel(level)
    logger.addHandler(handler)
    try:
        yield
    finally:
        logger.removeHandler(handler)
        logger.setLevel(original_level)


def _parse_command_line():
    """ Parse command line arguments for the main function.

    Returns
    -------
    namespace : argparse.Namespace
        Parsed command line arguments.
    """
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--debug',
        help="Log debugging information to console.",
        action="store_const",
        dest="log_level",
        const=logging.DEBUG,
        default=logging.WARNING,
    )
    parser.add_argument(
        '-v', '--verbose',
        help="Log information to console.",
        action="store_const",
        dest="log_level",
        const=logging.INFO,
    )
    return parser.parse_args()


def _create_demo(infos=None, title=_TITLE):
    """ Create the demo object with everything setup ready to be launched.

    Parameters
    ----------
    infos : list of dict, or None
        List of responses specifying the demo resources.
        Each response is a dictionary, in the format as specified by an
        entry point.
        If none, then responses are loaded from existing entry points installed
        in the Python environment.
    title : str, optional
        Default application title.

    Returns
    -------
    demo : Demo
    """
    if infos is None:
        infos = get_responses()

    resources = [
        response_to_node(response)
        for response in infos
    ]
    logger.info("Found %r resource(s).", len(resources))
    return Demo(
        title=title,
        model=DemoVirtualDirectory(resources=resources),
    )


def main(infos=None, title=_TITLE):
    """ Main function for launching the demo application.

    Parameters
    ----------
    infos : list of dict, or None
        List of responses specifying the demo resources.
        Each response is a dictionary, in the format as specified by an
        entry point. This allows packages to launch the demo application
        with their own set of data files without the entry points and without
        having to load files from other packages.
        If none, then responses are loaded from existing entry points installed
        in the Python environment.
    title : str, optional
        Default application title.
    """
    args = _parse_command_line()
    with _set_logger(logging.getLogger(_THIS_PACKAGE), level=args.log_level):
        demo = _create_demo(infos=infos, title=title)
        demo.configure_traits()
